use std::error::Error;

use bevy_ecs::world::World;

pub enum AppExit {
    Success,
    Error(Box<dyn Error>),
}

pub type RunnerFn = Box<dyn FnOnce(App) -> AppExit>;

#[derive(Debug, thiserror::Error)]
pub enum AppError {
    #[error("Runner is not set")]
    RunnerNotSet,
    #[error("Runner returned an error : {0}")]
    RunnerError(Box<dyn Error>),
}

pub struct App {
    world: World,
    runner: Option<RunnerFn>,
}

impl Default for App {
    fn default() -> Self {
        Self {
            world: World::new(),
            runner: None,
        }
    }
}

impl App {
    pub fn world_mut(&mut self) -> &mut World {
        &mut self.world
    }

    pub fn world(&self) -> &World {
        &self.world
    }

    pub fn run(mut self) -> Result<(), AppError> {
        match self.runner.take() {
            Some(runner) => match runner(self) {
                AppExit::Success => Ok(()),
                AppExit::Error(e) => Err(AppError::RunnerError(e)),
            },
            None => Err(AppError::RunnerNotSet),
        }
    }

    pub fn set_runner(&mut self, runner: RunnerFn) {
        self.runner = Some(runner);
    }
}
